/*
**++
**  FACILITY:
**      FEEDCHECK.C
**
**  ABSTRACT:
**      ANU NEWS Log File Analyzer
**
**  AUTHOR:
**      Mark Pizzolato (mark@infocomm.com)		3/27/92
**
**  COPYRIGHT:
**      Copyright  1992
**--
**/

#ifdef __TYPES
#undef __TYPES
#endif

#include <stdio.h>	/* Standard I/O Definitions			*/
#include <ctype.h>	/* Character Classifications			*/
#include <errno.h>	/* Error Codes					*/
#include <stat.h>	/* File Status Definitions			*/
#include <types.h>	/* Type Definitions				*/
#include <time.h>	/* Unix Time Definitions			*/
#include <descrip.h>	/* VMS Descriptor Definitions			*/
#include <clidef.h>	/* Comamnd Line Interpreter Codes		*/
#include <climsgdef.h>	/* Command Line Interpreter Message Codes	*/
#include <string.h>	/* String Handling Routine Declarations		*/
#include <stsdef.h>	/* Status Code Definitions			*/
#include <rms.h>	/* All RMS Definitions				*/

#define SS_FAILED(status)	(!((status)&STS$M_SUCCESS))
#define SS_OK(status)		(!SS_FAILED(status))

#define DLEN(x)  (x).dsc$w_length
#define DPTR(x)  (x).dsc$a_pointer

#define $DESCRIPTOR_D(x)  struct dsc$descriptor x = { 0, DSC$K_DTYPE_T, \
                          DSC$K_CLASS_D, 0 }

typedef struct {                  /* Quadword time structure */
   unsigned long int l0, l1;
   } uquad;

int rejects, junked, dups;

struct action_count
    {
    char *reason;
    int length;
    int *counter;
    } reasons[] = 
    {
    {"JUNK (SYS filter)", 				17,	&junked},
    {"JUNK (No local Newsgroup match)", 		31,	&junked},
    {"REJECT (SYS filter)", 				19,	&rejects},
    {"REJECT (No local Newsgroup match)", 		33,	&rejects},
    {"REJECT (Path Loop)", 				18,	&dups},
    {"REJECT (Duplicate - Cannot add Locally: Loop?)",	46, 	&dups},
    {"REJECT (/NOJUNK)", 				16,	&rejects},
    {"REJECT (No Message-ID)", 				22,	&rejects},
    {"REJECT (No JUNK)",		 		16,	&rejects},
    {"REJECT\n",		 			 7,	&rejects},
    {"REJECT (No Newsgroup)", 				21,	&rejects},
    {"JUNK (No Newsgroup)", 				19,	&junked},
    {"REJECT (",		 			 8,	&rejects},
    {"JUNK (",		 				 6,	&junked},
    {NULL,						 0, 	NULL}
    };

struct junk_rec
    {
    struct junk_rec *link;
    struct junk_host_rec *hosts;
    int junked;
    char newsgroup[256];
    };

struct junk_host_rec
    {
    struct junk_host_rec *link;
    int count;
    char hostname[80];
    };

struct junk_rec *garb = NULL;

int group_count = 0;

struct junk_rec **grouporder = NULL;

struct file_rec
    {
    struct file_rec *link;
    int lines;
    int articles;
    int dups;
    int junked;
    int rejects;
    int badlines;
    int batchsize;
    struct stat statb;
    char filename[L_tmpnam];
    char hostname[80];
    };

struct host_rec
    {
    struct host_rec *link;
    struct file_rec *files;
    int file_count;
    int articles;
    int dups;
    int junked;
    int rejects;
    int batchsize;
    int xmit_articles;
    int xmit_bytes;
    int xmit_article_ids;
    char hostname[80];
    };

struct host_rec *hosts = NULL;

struct host_rec **printorder = NULL;

/* Forward Declarations */
static int usage();
static int compare();
static int compareg();
static char *strtoupper();
static char *strtolower();
int *c$alloc_tmp();
struct dsc$descriptor *c$dsc();
struct dsc$descriptor *c$ldsc();
void lib$sig_to_ret();

globalvalue FEEDCHECKCMD;

main()
{
    char c;
    char *reason, *newsgroups, *p;
    char *opt_Hostname = "undetermined";
    char *opt_output = NULL;
    char Hostname[64] = "";
    FILE *Log;
    char filename[256];
    char bfilename[256];
    $DESCRIPTOR_D(fdesc);
    $DESCRIPTOR_D(rdesc);
    $DESCRIPTOR_D(lastdesc);
    int context, status, status_value, bfilesize;
    int verbose = 0;
    int record_junk = 0;
    int file_count = 0, host_count = 0;
    int tot_articles = 0, tot_dups = 0, tot_junked = 0, tot_rejected = 0;
    int tot_xmit_articles = 0, tot_xmit_bytes = 0, tot_xmit_article_ids = 0;
    int tot_batchsize = 0;
    int lines, articles, badlines;
    char logline[1024];
    struct action_count *a;
    struct stat statb;
    int beforetim[2] = {0xffffffff, 0x7fffffff};
    time_t beforetime = 0x7fffffff;
    int sincetim[2] = {0, 0};
    time_t sincetime = 0;
    time_t vms_to_unix_time();
    struct file_rec *fr;
    struct host_rec *hr;
    struct junk_rec *jr;
    struct junk_host_rec *jhr;

    lib$establish(lib$sig_to_ret);

    init_cli(FEEDCHECKCMD, "FEEDCHECK");

    cli_qualifier_action("HELP", usage);
    cli_get_time_value("BEFORE", beforetim);
    if (beforetim[1] != 0x7fffffff)
	beforetime = vms_to_unix_time(beforetim);
    cli_get_time_value("SINCE", sincetim);
    if ((sincetim[0]) || (sincetim[1]))
	sincetime = vms_to_unix_time(sincetim);
    cli_get_bool_value("JUNK", &record_junk);
    cli_get_int_value("VERBOSE", &verbose);
    cli_get_string_value("HOSTNAME", &opt_Hostname);
    strtolower(opt_Hostname);
    cli_get_string_value("OUTPUT", &opt_output);
    if (opt_output)
	if (NULL == freopen(opt_output, "w", stdout))
	    {
	    perror(opt_output);
	    exit(vaxc$errno);
	    }
    /* Assume the rest of the arguments are Logfiles to scan to report on */
    while (SS_OK(cli$get_value(c$dsc("FILENAME"), &fdesc)))
	{
	context = 0;
	while (SS_OK(lib$find_file(&fdesc, &rdesc, &context, 0, 
				   &lastdesc, &status_value, &0)))
	    {
	    strncpy(filename, DPTR(rdesc), DLEN(rdesc));
	    filename[DLEN(rdesc)] = '\0';
	    str$copy_dx(&lastdesc, &rdesc);
	    if (verbose>2)
		printf("Checking File: %s\n", filename);
	    if (stat(filename, &statb))
		{
		perror(filename);
		continue;
		}
	    if (statb.st_mtime < sincetime)
		continue;
	    if (statb.st_mtime > beforetime)
		continue;
	    if (NULL == (Log = fopen(filename, "r", "shr=get,put", "mbc=40")))
		{
		perror(filename);
		continue;
		}
	    ++file_count;
	    if (verbose>1)
		printf("Processing Log File: %s\n", filename);
	    fr = calloc(1, sizeof(*fr));
	    fr->statb = statb;
	    strcpy(fr->filename, filename);
	    rejects = junked = dups = 0;
	    strcpy(Hostname, "");
	    while (fgets(logline, sizeof(logline), Log))
		{
		++fr->lines;
		if (0 == memcmp(logline, "...Processing BATCH - h:", 24))
		    {
		    strcpy(Hostname, strtolower(strtok(&logline[24], " \t\n\r\f")));
		    continue;
		    }
		if (memcmp(logline, "Add ", 4))
		    {
		    char fhostname[80];
		    int fcount, fbytes, fids = 0;

		    if (memcmp(logline, "- Forwarded: ", 13))
			continue;
		    if (3 != sscanf(logline, "- Forwarded: %d/%d Articles/Bytes to site: %s",
				    &fcount, &fbytes, fhostname))
			if (2 != sscanf(logline, "- Forwarded: %d Article ID's to site: %s",
				    &fids, fhostname))
			    continue;
		    for (hr=hosts; hr; hr=hr->link)
			if (0 == strcmp(fhostname, hr->hostname))
			    break;
		    if (!hr)
			{
			++host_count;
			if (host_count == 1)
			    printorder = calloc(1, sizeof(printorder[0]));
			else
			    printorder = realloc(printorder, host_count*sizeof(printorder[0]));
			hr = calloc(1, sizeof(*hr));
			printorder[host_count-1] = hr;
			hr->link = hosts;
			hosts = hr;
			strcpy(hr->hostname, fhostname);
			}
		    if (fids)
			hr->xmit_article_ids += fids;
		    else
			{
			hr->xmit_articles += fcount;
			hr->xmit_bytes += fbytes;
			}
		    continue;
		    }
		if (2 == sscanf(logline, "Add - Reading file: %s (%d bytes)",
				bfilename, &bfilesize))
		    {
		    if (verbose>1)
			printf("Analyzing Log for Batch: %s\n", bfilename);
		    fr->batchsize += bfilesize;
		    continue;
		    }
		if (NULL == (newsgroups = strstr(logline, "> ")))
		    {
		    ++fr->badlines;
		    continue;
		    }
		else
		    {
		    char linehost[80];

		    if (1 == sscanf(logline, "Add %s <", linehost))
			{
			if (strcmp(Hostname, linehost))
			    {
			    if ((Hostname[0] != '\0') && (verbose>1))
				printf("Log Host: %s, Batch Host: %s\n",
					Hostname, linehost);
			    strcpy(Hostname, linehost);
			    }
			if (fr->hostname[0] == '\0')
			    strcpy(fr->hostname, linehost);
			else
			    if (strcmp(fr->hostname, linehost))
				printf("Unexpected source host change from %s to %s in file: %s\n",
					fr->hostname, linehost, fr->filename);
			}
		    }
		newsgroups += 2;
		++fr->articles;
		if (NULL == (reason = strstr(logline, ": ")))
		    continue;
		reason += 2;
		for (a=reasons; a->reason; ++a)
		    if (0 == memcmp(a->reason, reason, a->length))
			{
			++*a->counter;
			if ((record_junk) && (a->counter == &junked))
			    {
			    *(reason-2) = '\0';
			    while (1)
				{
				if (p = strchr(newsgroups, ','))
				    *p = '\0';
				for (jr=garb; jr; jr=jr->link)
				    if (0 == strcmp(newsgroups, jr->newsgroup))
					break;
				if (!jr)
				    {
				    ++group_count;
				    if (group_count == 1)
					grouporder = calloc(1, sizeof(grouporder[0]));
				    else
					grouporder = realloc(grouporder, group_count*sizeof(grouporder[0]));
				    jr = calloc(1, sizeof(*jr));
				    grouporder[group_count-1] = jr;
				    jr->link = garb;
				    garb = jr;
				    strcpy(jr->newsgroup, newsgroups);
				    }
				++jr->junked;
				for (jhr=jr->hosts; jhr; jhr=jhr->link)
				    if (0 == strcmp(jhr->hostname, Hostname))
					break;
				if (!jhr)
				    {
				    jhr = calloc(1, sizeof(*jhr));
				    strcpy(jhr->hostname, Hostname);
				    jhr->link = jr->hosts;
				    jr->hosts = jhr;
				    }
				++jhr->count;
				newsgroups += (1 + strlen(newsgroups));
				if (isspace(*newsgroups))
				    break;
				}
			    }
			break;
			}
		}
	    fclose(Log);
	    if (!fr->articles)
		{
		free(fr);
		continue;
		}
	    fr->rejects = rejects;
	    fr->junked = junked;
	    fr->dups = dups;
	    if (0 == strcmp(Hostname, ""))
		strcpy(Hostname, fr->hostname);
	    for (hr=hosts; hr; hr=hr->link)
		if (0 == strcmp(Hostname, hr->hostname))
		    break;
	    if (!hr)
		{
		++host_count;
		if (host_count == 1)
		    printorder = calloc(1, sizeof(printorder[0]));
		else
		    printorder = realloc(printorder, host_count*sizeof(printorder[0]));
		hr = calloc(1, sizeof(*hr));
		printorder[host_count-1] = hr;
		hr->link = hosts;
		hosts = hr;
		strcpy(hr->hostname, Hostname);
		}
	    ++hr->file_count;
	    hr->articles += fr->articles;
	    hr->dups += fr->dups;
	    hr->junked += fr->junked;
	    hr->rejects += fr->rejects;
	    hr->batchsize += fr->batchsize;
	    fr->link = hr->files;
	    hr->files = fr;
	    tot_articles += fr->articles;
	    tot_dups += fr->dups;
	    tot_junked += fr->junked;
	    tot_rejected += fr->rejects;
	    tot_batchsize += fr->batchsize;
	    Hostname[0] = '\0';
	    }
	lib$find_file_end(&context);
	}
    printf("Site:     Files:Accept: Dups:Junked:Reject:Total: %%Bad:%%TAcpt: I've: Xmit:Bytes:\n");
    printf("================================================================================\n");
    if (host_count > 1)
	{
	int i;

	qsort((char *)printorder, host_count, sizeof(printorder[0]), compare);
	hosts = printorder[0];
	printorder[host_count-1]->link = NULL;
	for (i=0; i<(host_count-1); ++i)
	    printorder[i]->link = printorder[i+1];
	}
    for (hr=hosts; hr; hr = hr->link)
	{
	tot_xmit_articles += hr->xmit_articles;
	tot_xmit_bytes += hr->xmit_bytes;
	tot_xmit_article_ids += hr->xmit_article_ids;
	for (fr = hr->files; fr; fr = fr->link)
	    {
	    if (verbose>1)
		printf("Log File: %s, Lines: %d, badlines: %d\n", 
			fr->filename, fr->lines, fr->badlines);
	    if (verbose)
		printf("%-14s%13d  %6d  %6d %8d\n",
			hr->hostname, fr->articles, fr->dups, fr->junked, fr->rejects);
	    }
	if ((hr->file_count > 0) || (hr->xmit_articles))
	    printline(hr->hostname, hr->file_count, hr->batchsize, 
		      hr->articles, hr->dups, hr->junked, hr->rejects, 
		      tot_articles-(tot_dups+tot_junked+tot_rejected), 
		      hr->xmit_articles, hr->xmit_bytes, hr->xmit_article_ids);
	}
    if (host_count > 1)
	{
	printf("\n          Files:Accept: Dups:Junked:Reject:Total: %%Bad:%%TAcpt: I've: Xmit:Bytes:\n");
	printline("TOTAL:", file_count, tot_batchsize, tot_articles, tot_dups, 
			    tot_junked, tot_rejected, 0, tot_xmit_articles,
		            tot_xmit_bytes, tot_xmit_article_ids);
	}
    if (group_count > 1)
	{
	int i;

	printf("\nJunked Newsgroups:\n");
	printf("======================\n");
	qsort((char *)grouporder, group_count, sizeof(grouporder[0]), compareg);
	garb = grouporder[0];
	grouporder[group_count-1]->link = NULL;
	for (i=0; i<(group_count-1); ++i)
	    grouporder[i]->link = grouporder[i+1];
	for (jr=garb; jr; jr=jr->link)
	    {
	    printf("%5d Articles in: %s\n", jr->junked, jr->newsgroup);
	    for (jhr=jr->hosts; jhr; jhr=jhr->link)
		printf("\t%4d from: %s\n", jhr->count, jhr->hostname);
	    }
	}
    return;
}

static usage()
{
    fprintf(stderr, "\n\tFEEDCHECK\n\n");
    fprintf(stderr, "\tProcesses the log files produced during ANU-NEWS maintenance\n");
    fprintf(stderr, "\toperations and produces a report of the activities observed.\n");
    fprintf(stderr, "\n\tFormat\n");
    fprintf(stderr, "\t   FEEDCHECK filespec[,...]\n");
    fprintf(stderr, "\n\tPARAMETER\n");
    fprintf(stderr, "\t\t\tfilespec[,...]\n");
    fprintf(stderr, "\t\t\tSpecifies one or more files to be processed.\n");
    fprintf(stderr, "\n\tQUALIFIERS\n");
    fprintf(stderr, "\t\t/HOSTNAME=host\n");
    fprintf(stderr, "\t\t\tSpecifies the default host to attribute traffic\n");
    fprintf(stderr, "\t\t\tto if the input file does not contain a host\n");
    fprintf(stderr, "\t\t\tspecifier.\n");
    fprintf(stderr, "\t\t/BEFORE[=time]\n");
    fprintf(stderr, "\t\t\tSpecifies that files created before the specified\n");
    fprintf(stderr, "\t\t\ttime are to be processed.\n");
    fprintf(stderr, "\t\t/SINCE[=time]\n");
    fprintf(stderr, "\t\t\tSpecifies that files created after the specified\n");
    fprintf(stderr, "\t\t\ttime are to be processed.\n");
    fprintf(stderr, "\t\t/JUNK\n");
    fprintf(stderr, "\t\t/NOJUNK (default)\n");
    fprintf(stderr, "\t\t\tCauses a detailed summary of articles that have been\n");
    fprintf(stderr, "\t\t\tjunked, and where they were fed from.\n");
    fprintf(stderr, "\t\t/OUTPUT[=filespec]\n");
    fprintf(stderr, "\t\t\tControls where the output of the command is sent.\n");
    fprintf(stderr, "\t\t\tBy default, the display is written to the current\n");
    fprintf(stderr, "\t\t\tSYS$OUTPUT device.\n");
    fprintf(stderr, "\t\t/VERBOSE[=n]\n");
    fprintf(stderr, "\t\t\tSelects a more verbose output format at level n.\n");
    fprintf(stderr, "\n");
}

printline(name, files, inbytes, articles, dups, junked, rejects, tot, xmit, bytes, ihave)
char *name;
int files, inbytes, articles, dups, junked, rejects, tot, xmit, bytes, ihave;
    {
    char files_bytes[40] = ". ";
    char buf[256];
    int accept;
    float kbytes, mbytes;

    if (files)
	{
	sprintf(files_bytes, "%d", files);
	kbytes = ((float)inbytes)/1024.0;
	mbytes = ((float)inbytes)/(1024.0*1024.0);
	if (kbytes > 1.0)
	    if (mbytes > 1.0)
	        sprintf(&files_bytes[strlen(files_bytes)], "(%0.1fM)", mbytes);
	    else
	        sprintf(&files_bytes[strlen(files_bytes)], "(%0.1fK)", kbytes);
	else
	    strcat(files_bytes, " ");
	}
    sprintf(buf, "%s%*s", name, 17-strlen(name), files_bytes);
    accept = articles - (dups + junked + rejects);
    if (accept)
	sprintf(&buf[strlen(buf)], "%5d", accept);
    else
	strcat(buf, "    .");
    if (dups)
	sprintf(&buf[strlen(buf)], "%6d", dups);
    else
	strcat(buf, "     .");
    if (junked)
	sprintf(&buf[strlen(buf)], "%7d", junked);
    else
	strcat(buf, "      .");
    if (rejects)
	sprintf(&buf[strlen(buf)], "%6d", rejects);
    else
	strcat(buf, "     .");
    if (articles)
	sprintf(&buf[strlen(buf)], "%7d", articles);
    else
	strcat(buf, "      .");
    if (dups+junked+rejects)
	sprintf(&buf[strlen(buf)], "%6.1f%%", 
			  ((float)(100*(dups+junked+rejects)))/articles);
    else
	strcat(buf, "      .");
    if ((tot) && (accept))
	sprintf(&buf[strlen(buf)], "%6.1f%%",
			  ((float)(100*accept))/(float)tot);
    else
	strcat(buf, "      .");
    if (ihave)
	sprintf(&buf[strlen(buf)], "%6d", ihave);
    else
	strcat(buf, "     .");
    if (xmit)
	sprintf(&buf[strlen(buf)], "%6d", xmit);
    else
	strcat(buf, "     .");
    kbytes = ((float)bytes)/1024.0;
    mbytes = ((float)bytes)/(1024.0*1024.0);
    if (kbytes > 1.0)
	if (mbytes > 1.0)
	    sprintf(&buf[strlen(buf)], "%5.1fM", mbytes);
	else
	    if (kbytes > 100.0)
		sprintf(&buf[strlen(buf)], "%5.0fK", kbytes);
	    else
		sprintf(&buf[strlen(buf)], "%5.1fK", kbytes);
    else
	if (bytes)
	    sprintf(&buf[strlen(buf)], "%6d", bytes);
	else
	    strcat(buf, "     .");
    printf("%s\n", buf);
    }

static int
compare(a, b)
struct host_rec **a, **b;
{
    return(strcmp((*a)->hostname, (*b)->hostname));
}

static int
compareg(a, b)
struct junk_rec **a, **b;
{
    return(strcmp((*a)->newsgroup, (*b)->newsgroup));
}

static char *
strtoupper(str)
char *str;
   {
   char *s = str;

   while (*s)
	{
	if (islower(*s))
	    *s = toupper(*s);
	++s;
	}
   return(str);
   }

static char *
strtolower(str)
char *str;
   {
   char *s = str;

   while (*s)
	{
	if (isupper(*s))
	    *s = tolower(*s);
	++s;
	}
   return(str);
   }

int *c$_tmphead = (int *) 0;

/************************************************************************/

int *c$alloc_tmp()
{
   int *tmp;
   
   tmp = (int *) malloc(12);
   *tmp = c$_tmphead;
   c$_tmphead = tmp;
   return(c$_tmphead + 1);
}

/************************************************************************/

struct dsc$descriptor *c$dsc(c$_string)
   char *c$_string;
{
   int *tmp;
   struct dsc$descriptor *c$_tmpdesc;

   while (c$_tmphead) {
      tmp = *c$_tmphead;
      free(c$_tmphead);
      c$_tmphead = tmp;
      }                     
   c$_tmpdesc = (struct dsc$descriptor *) c$alloc_tmp();
   c$_tmpdesc->dsc$w_length = strlen(c$_string);
   c$_tmpdesc->dsc$b_dtype = DSC$K_DTYPE_T;
   c$_tmpdesc->dsc$b_class = DSC$K_CLASS_S;
   c$_tmpdesc->dsc$a_pointer = c$_string;
   return(c$_tmpdesc);
}

/************************************************************************/

struct dsc$descriptor *c$ldsc(c$_string,c$_length)
   char *c$_string;
   int c$_length;
{
   int *tmp;
   struct dsc$descriptor *c$_tmpdesc;

   while (c$_tmphead) {
      tmp = *c$_tmphead;
      free(c$_tmphead); 
      c$_tmphead = tmp;
      }
   c$_tmpdesc = (struct dsc$descriptor *) c$alloc_tmp();
   c$_tmpdesc->dsc$w_length = c$_length;
   c$_tmpdesc->dsc$b_dtype = DSC$K_DTYPE_T;
   c$_tmpdesc->dsc$b_class = DSC$K_CLASS_S;
   c$_tmpdesc->dsc$a_pointer = c$_string;
   return(c$_tmpdesc);
}

/************************************************************************/

void cksts(status)
    int status;
{
    if (!(status & STS$M_SUCCESS))
	lib$signal(status);
}

cli_qualifier_action(qualifier,action_routine)
   char *qualifier;
   int (*action_routine)();
{
    int value;

    cli_get_bool_value(qualifier, &value);
    if (value)
	(*action_routine)();
}

cli_get_bool_value(qualifier,value)
   char *qualifier;
   int *value;
{
    unsigned long int status;

    switch(status = cli$present(c$dsc(qualifier)))
	{
	case CLI$_PRESENT:
	case CLI$_LOCPRES:
	    *value = 1;
	    break;   
	case CLI$_ABSENT:
	case CLI$_NEGATED:
	case CLI$_LOCNEG:
	    *value = 0;
	    break;   
	}
}

cli_get_int_value(qualifier,value)
   char *qualifier;
   int *value;
{
    unsigned long int status;
    $DESCRIPTOR_D(rdesc);
    extern int ots$cvt_tz_l();
    extern int ots$cvt_tu_l();
    extern int ots$cvt_to_l();
    int (*rtn)();
    long int num;
    char *c;

    if (cli$present(c$dsc(qualifier)) == CLI$_PRESENT)
	{
	cksts(status = cli$get_value(c$dsc(qualifier),&rdesc));
	rtn = ots$cvt_tu_l;
	if ((DLEN(rdesc) > 2) && (rdesc.dsc$a_pointer[0] == '%'))
	    {
	    if (rdesc.dsc$a_pointer[1] == 'X')
		rtn = ots$cvt_tz_l;
	    if (rdesc.dsc$a_pointer[1] == 'O')
		rtn = ots$cvt_to_l;
	    }
	cksts(status = (*rtn)(&rdesc, &num, 4, 0));
	*value = num;
	}
}

cli_get_string_value(qualifier,value)
   char *qualifier;
   char **value;
{
    unsigned long int status;
    $DESCRIPTOR_D(rdesc);
    $DESCRIPTOR_D(vdesc);
    long int num;

    if (cli$present(c$dsc(qualifier)) == CLI$_PRESENT)
	{
	cksts(status = cli$get_value(c$dsc(qualifier),&rdesc));
	num = 1 + DLEN(rdesc);
	str$copy_r(&vdesc, &num, DPTR(rdesc));
	*value = DPTR(vdesc);
	(*value)[DLEN(rdesc)] = '\0';
	str$free1_dx(&rdesc);
	}
}

cli_get_time_value(qualifier,value)
   char *qualifier;
   uquad *value;
{
    char tmp1[64], tmp2[64];
    unsigned long int status;
    $DESCRIPTOR_D(rdesc);
    uquad restim = { 0, 0 };
   
    if (cli$present(c$dsc(qualifier)) == CLI$_PRESENT)
	{
	status = cli$get_value(c$dsc(qualifier),&rdesc);
	if (status == CLI$_ABSENT)
	    cksts(sys$gettim(&restim));
	else
	    {
	    cksts(status);
	    strncpy(tmp1,DPTR(rdesc),DLEN(rdesc));
	    tmp1[DLEN(rdesc)] = '\0';
	    str$free1_dx(&rdesc);
	    if (!strcmp(tmp1,"TODAY"))
		{
		cksts(sys$asctim(0,c$ldsc(tmp2,sizeof(tmp2)),0,0));
		*(strchr(tmp2,':')-3) = '\0';
		sprintf(tmp1,"%s 00:00:00.00",tmp2);
		}
	    cksts(sys$bintim(c$dsc(tmp1),&restim));
	    }
	*value = restim;
	}
}

init_cli(char *table, char *name)
{
    $DESCRIPTOR_D(cmd);
    int i, *lib$get_input();
  
    cksts(cli$get_value(c$dsc("$VERB"),&cmd));

    if (memcmp(DPTR(cmd),name,DLEN(cmd)) == 0)
	return(1);    /* the command must have been properly defined! */
    /* this code assumes that the verb is shorter than than the foreign
	symbol (which includes device:[dir], so, should be reasonable) */
    if ((DLEN(cmd) == 3) && 
	(0 == memcmp("RUN",DPTR(cmd),3)))
	{
	i = strlen(name);
	str$copy_r(&cmd, &i, name);
	}
    else
	{
	cksts(cli$get_value(c$dsc("$LINE"),&cmd));

	for (i = 0; (i < DLEN(cmd)) &&
		    (cmd.dsc$a_pointer[i] != ' ') &&
		    (cmd.dsc$a_pointer[i] != '/'); ++i)
	    cmd.dsc$a_pointer[i] = (i < strlen(name))?name[i]:' ';
	}

    cksts(cli$dcl_parse(&cmd,table,lib$get_input,lib$get_input));
}

time_t vms_to_unix_time(time)
int time[2];
{
static int ubase[2] = {0x4beb4000,0x007c9567}; /* Unix Time Base 1/1/70 */
int     utime[2];
time_t  result;
int     divisor = 10000000;
int     remainder;

    lib$subx(time, ubase, utime);
    lib$ediv(&divisor, utime, &result, &remainder);
    return(result);
}
