$!++
$!  FACILITY:
$!      NEWSBUILD
$!
$!  AUTHOR:
$!      Geoff Huston
$!      Copyright  1988,1989,1990,1991,1992
$!
$!  FUNCTIONALITY:
$!      This file can be used to rebuild the NEWS executables either from
$!      scratch, or can be used for an incremental build if one (or a number)
$!      of source files have been edited. Both the production and debug
$!      executables are constructed in this procedure.
$!
$!  PRECONDITIONS:
$!      - All sources are in Dev:[NEWS.NEWS_SRC]
$!      - The default directory is set to DEV:[NEWS.NEWS_BUILD]
$!      - Execute by "$ @[-.news_src]newsbuild"
$!
$!  Version:
$!      V5.5     7-Oct-1988     GIH
$!        - Included modifications for WIN TCP/IP V3.0 support by Jim Patterson
$!          (jimp@cognos.uucp).
$!          NOTE: I have NOT tested the WIN support modifications
$!      V5.6    11-Nov-1988     GIH
$!        - Version 5.6 configuration
$!      V5.8     2-Mar-1989     GIH
$!        - Add NEWSV58_UPGRADE, NEWSDIST, NEWSRTL modules
$!	V5.9	14-Jul-1989	GIH
$!	  - Reduce the file activity to speed up the routine
$!	  - now included in NEWS_SRC directory
$!	V6.0	13-Jun-1989	GIH
$!	  - use parameter 1 to specify the TCP transport (or ask if not supplied)
$!	V6.0-1	 3-Nov-1990	(GIH added) - James Gerland: TWG changes
$!	V6.0-4	25-May-1991	volz@process.com
$!	  - Added TCPware support
$!	V6.1	31-Jan-1991	ake@dayton.saic.com Earle Ake
$!	  - Dont deassign newsrc logical if it isn't defined
$!	V6.1	 3-Feb-1991	ake@dayton.saic.com Earle Ake
$!	  - For Multinet define "__TYPES" to avoid compiler noise
$!      v6.1     9-Feb-1991     don@beckman.com Don Roberts
$!        - Allow compile only.  Necessary because my site has the c compiler
$!          on a different node from TCP/IP.  Also if you define the logical
$!          env_verify to 1, verification is set on
$!      v6.1     17-Feb-1991     mark@infocomm.com Mark Pizzolato
$!	  - Let Control-Y during a compile actually stop compiling and not 
$!	    simply abort the current compile and go on to the next.
$!
$!      v6.1     03-Mar-1993     bailey@genetics.upenn.edu
$!        -  p6 = purgetype.  If p6 contains string NOOBJ, object files will
$!           not be purged.  If p6 contains string NOEXE, images will not be
$!           purged.
$!
$!	V6.1b7	13-Jul-1993	bailey@genetics.upenn.edu  Charles Bailey
$!			  - changed WIN/TCP link option to use shareable image library
$!			    for PathWay 1.0 if it exists
$!	V6.1b7  20-May-1993  Charles Bailey  bailey@genetics.upenn.edu
$!	  - fixed minor bug in makeobj subroutine
$!	  - fixed bug in News.Exe and NewsDbg.Exe generation for link-only runs
$!	V6.1b9  29-Dec-1993  Mark Pizzolato mark@infocomm.com
$!	  - Moved NNTP_TTY.EXE to [.NEWS_BUILD] and made sure that it is always
$!	    built with the debug compiled pieces so that it can serve it's
$!	    function as a test framework for debugging the NNTP_SERVER code.
$!	V6.1b8  20-Oct-1993  Bernie Volz  volz@process.com
$!    - updated TCPware logical names
$!
$!  Usage:
$!	$ SET DEF [.NEWS_SRC]
$!	$ @NEWSBUILD [tcp-transport] [force] [debug] [client-only] [compile-defines]
$!
$!		tcp-transport:  N [ONE]
$!			        M [ULTINET]
$!			        C [MU]
$!				W [IN]
$!				U [CXQIOS]
$!				UCXS [OCKETS]
$!				E [XOS]]
$!				T [CPware]
$!
$!		force:		COMPILE only - force full compilation and skip link
$!				LINK only - skip compilation and force full link
$!                              BOTH compile and link - force full compile and full link
$!				Objects - compile only - but only based on date/time check and skip link
$!				Exes - skip compilation and link - but only based on date/time check
$!              		<empty string> - both compile and link, but only as required by date/time check
$!
$!		debug:		DEBUG   - compile and link with DEBUG option
$!				NODEBUG - compile and link with NODEBUG option
$!                              <empty string> - Do both
$!
$!		client-only:	CLIENT - build NEWS.EXE and NEWS.HLB only
$!				<empty string> - build everything
$!
$!		compile-defines:
$!				<define list> - comma separated list of 
$!						identifiers/value pairs which 
$!						will be passed to compiles of
$!						C code.
$!				<empty string> - no extra compile time defines
$!
$!--
$!
$ makehay_ver='f$verify(0)
$ image_ver == f$environment ("verify_image")   !store verification info
$ proc_ver == 'f$verify (f$trnlnm("ENV_VERIFY"), image_ver)
$ axpflag = F$GetSyi("HW_MODEL") .gt. 1024  ! Probably ALPHA
$ newsrclog=f$trnlnm("newsrc")
$ if newsrclog .nes. "" then $ deassign newsrc
$ set def [-.news_build]
$ on control_y THEN $ stop
$ on error THEN $ goto done
$ if axpflag then define lnk$library sys$library:vaxcrtl
$ define vaxc$include  sys$library
$ if axpflag
$ then
$   define link_opt [-.news_src]options_c_link_axp.opt
$ else
$   define link_opt [-.news_src]options_c_link.opt
$ endif
$ defsw = ""
$ use_multinet = 0
$ use_cmu = 0
$ use_twg = 0
$ use_ucxqios = 0
$ use_ucxsockets = 0
$ use_exos = 0
$ use_tcpware = 0
$ IF p1 .eqs. ""
$ THEN
$   write sys$output "TCP transport options:"
$   write sys$output "      NONE"
$   write sys$output "      MULTINET"
$   write sys$output "      CMU"
$   write sys$output "      WIN"
$   write sys$output "      UCXQIOS"
$   write sys$output "      UCXSOCKETS (not tested thoroughly)"
$   write sys$output "      EXOS"
$   write sys$output "      TCPware"
$   inquire p1 "TCP Transport [NONE]"
$ ENDIF
$ tcptransport = f$extract(0,1,p1)
$ IF tcptransport .EQS. "M" THEN $ use_multinet = 1
$ IF tcptransport .EQS. "C" THEN $ use_cmu = 1
$ IF tcptransport .EQS. "W" THEN $ use_twg = 1
$ IF tcptransport .EQS. "E" THEN $ use_exos = 1
$ IF tcptransport .EQS. "T" THEN $ use_tcpware = 1
$ IF tcptransport .EQS. "U"
$ THEN
$   tcptransport = f$extract(0,4,p1)
$   IF tcptransport .EQS. "UCXS"
$   THEN
$     use_ucxsockets = 1
$   ELSE
$     use_ucxqios = 1
$   ENDIF
$ ENDIF
$!
$! Call in CMUTCP support
$!
$ IF use_cmu
$ THEN
$   defsw = "/DEFINE=(CMU=1)"
$   write sys$output "CMU IP support included"
$ ENDIF
$!
$! Call in UCXSOCKETS support
$!
$ IF use_ucxsockets
$ THEN
$   write sys$output "ANU-NEWS under UCXSOCKETS is not thoroughly tested,"
$   write sys$output "the use of UCXQIOS is recommended.  Also the"
$   write sys$output "NNTP_TCPUCXM server is not compiled under UCXSOCKETS."
$   IF f$search("sys$library:ucx$inetdef.h") .EQS. ""
$   THEN
$     write sys$output "UCX support cannot be loaded - no library file"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(UCXSOCKETS=1,UCX=1)"
$   if axpflag
$   then
$     DEFINE link_opt [-.news_src]options_c_ucxlink_axp.opt
$   else
$     DEFINE link_opt [-.news_src]options_c_ucxlink.opt
$   endif
$   write sys$output "UCX IP support included (socket drivers)"
$ ENDIF
$!
$! Call in UCXQIOS support
$!
$ IF use_ucxqios
$ THEN
$   IF f$search("sys$library:ucx$inetdef.h") .EQS. ""
$   THEN
$     write sys$output "UCX support cannot be loaded - no library file"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(UCXQIOS=1,UCX=1)"
$   if axpflag
$   then
$     DEFINE link_opt [-.news_src]options_c_ucxlink_axp.opt
$   else
$     DEFINE link_opt [-.news_src]options_c_ucxlink.opt
$   endif
$   write sys$output "UCX IP support included (qio drivers)"
$ ENDIF
$!
$! Call in Multinet support based on multinet include files
$!
$ IF use_multinet
$ THEN
$   IF f$search("multinet_root:[multinet.include]*.h") .EQS. ""
$   THEN
$     write sys$output "MULTINET support cannot be loaded - no library files"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(MULTINET=1,__TYPES=1)"
$   DEFINE vaxc$include  multinet_root:[multinet.include],-
			   multinet_root:[multinet.include.sys],-
			   multinet_root:[multinet.include.arpa],-
			   multinet_root:[multinet.include.netinet],-
			   multinet_root:[multinet.include.vms],-
                      	   sys$library
$   if axpflag
$   then
$     DEFINE VMS MULTINET_ROOT:[MULTINET.INCLUDE.VMS]
$     DEFINE ARPA MULTINET_ROOT:[MULTINET.INCLUDE.ARPA]
$     DEFINE SYS MULTINET_ROOT:[MULTINET.INCLUDE.SYS]
$     DEFINE NETINET MULTINET_ROOT:[MULTINET.INCLUDE.NETINET]
$     DEFINE link_opt [-.news_src]options_c_multinetlink_axp.opt
$   else
$     DEFINE link_opt [-.news_src]options_c_multinetlink.opt
$   endif
$   write sys$output "Multinet IP support included"
$ ENDIF
$!
$! Call in TCPware support based on TCPware include files
$!
$ IF use_tcpware
$ THEN
$   IF f$search("TCPWARE_INCLUDE:*.h") .EQS. ""
$   THEN
$     write sys$output "TCPware support cannot be loaded - no library files"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(TCPWARE=1)"
$   DEFINE vaxc$include    tcpware_include,-
                           sys$library
$   if axpflag
$   then
$     DEFINE link_opt [-.news_src]options_c_tcpwarelink_axp.opt
$   else
$     DEFINE link_opt [-.news_src]options_c_tcpwarelink.opt
$   endif
$   write sys$output "TCPware TCP/IP support included"
$ ENDIF
$!
$!
$! Call in TWG support based on TWG include files
$!
$ IF use_twg
$ THEN
$   IF f$search("TWG$TCP:[NETDIST.INCLUDE]*.H") .eqs. ""
$   THEN
$     write sys$output "WIN support cannot be loaded - no library files"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(TWG=1)"
$   define vaxc$include sys$library,twg$tcp:[netdist.include],-
                          twg$tcp:[netdist.include.sys],-
                          twg$tcp:[netdist.include.vms],-
                          twg$tcp:[netdist.include.netinet],-
                          twg$tcp:[netdist.include.arpa],-
                          sys$library
$   define sys twg$tcp:[netdist.include.sys]
$   If F$Search("Sys$Share:TWGLib.Exe").nes.""
$   Then
$     if axpflag
$     then
$       define link_opt [-.news_src]options_c_winlink_axp.opt
$     else
$       define link_opt [-.news_src]options_c_winlink.opt
$     endif
$     write sys$output "TWG IP support included (shareable image library)"
$   Else
$     define lnk$library twg$tcp:[netdist.lib]twglib
$     define lnk$library_1 sys$library:vaxcrtl
$     write sys$output "TWG IP support included (object library)"
$   EndIf
$ ENDIF
$ IF use_exos
$ THEN
$   IF f$trnlnm("EXOS_ETC").eqs.""
$   THEN
$     write sys$output "EXOS not found!"
$     exit
$   ENDIF
$   defsw = "/DEFINE=(EXOS=1)"
$   if axpflag
$   then
$     define link_opt [-.news_src]options_c_link_axp.opt
$   else
$     define link_opt [-.news_src]options_c_link.opt
$   endif
$ ENDIF
$!
$!	Examine parameter 2 (Compile | Link | Both | Objects | Exes | empty)
$!
$ force = f$extract(0,1,p2)
$ skip_compile = 0
$ skip_link = 0
$ force_compile = 0
$ force_link = 0
$ IF force .EQS. "O" THEN $ skip_link = 1
$ IF force .EQS. "E" THEN $ skip_compile = 1
$ IF force .EQS. "C"
$ THEN
$  force_compile = 1
$  skip_link = 1
$ ENDIF
$ IF force .EQS. "L"
$ THEN 
$  force_link = 1
$  skip_compile = 1
$ ENDIF
$ IF force .EQS. "B"
$ THEN
$  force_compile = 1
$  force_link = 1
$ ENDIF
$!
$!	Examine parameter 3 (DEBUG | NODEBUG | empty)
$!
$ dbgtype = f$extract(0,1,p3)
$ make_debug = 1
$ make_nodebug = 1
$ IF dbgtype .EQS. "D" THEN $ make_nodebug = 0
$ IF dbgtype .EQS. "N" THEN $ make_debug = 0
$!	Examine parameter 4 (Client-only | empty)
$!
$!	client-only means build NEWS.EXE and NEWS.HLB only; ignore the rest
$ client_only = ( f$edit(f$extract(0,1,p4),"UPCASE") .eqs. "C" )
$!	Examine parameter 5 (Extra Compile Defines | empty)
$!
$ IF p5.nes.""
$ THEN
$  IF defsw.eqs.""
$  THEN
$   defsw = "/DEFINE=("+p5+")"
$  ELSE
$   defsw = f$extract(0,f$length(defsw)-1,defsw)+p5+")"
$  ENDIF
$ ENDIF
$!
$!
$ IF axpflag	! Probably ALPHA
$ THEN
$	IF USE_MULTINET
$	THEN
$		CC_AXP_SW = "/PREFIX=(ALL_ENTRIES,EXCEPT=(SOCKET,CONNECT,BIND,LISTEN,SOCKET_READ,SOCKET_WRITE,SOCKET_CLOSE,SELECT,ACCEPT,BCMP,BCOPY,BZERO,GETHOSTBYNAME,"
$		CC_AXP_SW = CC_AXP_SW + "GETSOCKOPT,SETSOCKOPT,GETHOSTBYADDR,GETPEERNAME,GETDTABLESIZE,HTONS,HTONL,NTOHS,NTOHL,SEND,SENDTO,RECV,RECVFROM))"
$	ELSE    CC_AXP_SW = ""
		
$	ENDIF
$	cc :== "cc''defsw'/NOMEMBER_ALIGN/STANDARD=VAXC/TIE''CC_AXP_SW'"
$ ELSE
$	cc :== "cc''defsw'"
$ ENDIF
$ show symb cc
$!
$! Save the Compile Time "Defines" for this compile.
$!
$ open/write temp [-.news_build]compile.defines
$ write temp defsw
$ close temp
$ if f$search("[-.news_build]compile.defines;-1").eqs.""
$ then
$   force_compile = 1
$   force_link = 1
$ else
$   differences/output=nla0: [-.news_build]compile.defines
$   if $severity.eq.3
$     then
$	write sys$output ""
$	if f$file_attributes("[-.news_build]compile.defines;-1","FFB").eqs.4
$	  then
$	    write sys$output "Forcing a complete recompile due to specification compile time defines"
$	  else
$	    write sys$output "Forcing a complete recompile due to compile defines changing from:"
$	    type [-.news_build]compile.defines;-1
$	endif
$	if f$length(defsw).eqs.0
$	then
$	  write sys$output "to compiling without any compile time defines."
$	else
$	  write sys$output "To:
$	  type [-.news_build]compile.defines
$	endif
$	write sys$output ""
$       force_compile = 1
$       force_link = 1
$   endif
$   purge [-.news_build]compile.defines
$ endif
$!
$!	Examine parameter 6 (Purge object,exe | do not purge)
$ purge_objs = 1
$ purge_exes = 1
$ If p6.nes.""
$ Then
$   p6 = F$Edit(p6,"Upcase,Collapse")
$   If F$Locate("NOOBJ",p6).lt.F$Length(p6) Then purge_objs = 0
$   If F$Locate("NOEXE",p6).lt.F$Length(p6) Then purge_exes = 0
$ EndIf
$!
$!  INPATHS.EXE - done first as there is no tcp code required
$!
$ IF f$search("[-.NEWS_SRC]INPATHS.C") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("INPATHS.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]INPATHS.C","rdt")).gts.f$cvtime(f$file_attributes("INPATHS.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc inpaths"
$     cc/nolis/object=[-.news_build]inpaths inpaths
$     set def [-.news_build]
$     if purge_objs then purge inpaths.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("INPATHS.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]INPATHS.EXE") .eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("INPATHS.OBJ","rdt")).gts.-
                  f$cvtime(f$file_attributes("[-.NEWS_DIST]INPATHS.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link inpaths"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]inpaths.exe inpaths,link_opt/opt
$     if purge_exes then purge [-.news_dist]inpaths.exe
$   ENDIF
$ ENDIF
$!
$! .OLB files
$!
$ IF f$search("NEWS_LIBRARY.OLB").eqs."" THEN $ library/create/obj news_library
$ IF f$search("NEWS_LIBRARYD.OLB").eqs."" THEN $ library/create/obj news_libraryd
$!
$!  get cre date time for include files
$!
$ vartime = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSVARIABLES.H","rdt"))
$ exttime = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSEXTERN.H","rdt"))
$ deftime = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSDEFINE.H","rdt"))
$ inctime = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSINCLUDE.H","rdt"))
$ sittime = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSSITE.H","rdt"))
$ call compile "NEWSADD"        "DELOBJ"
$ call compile "NEWSCALLEDIT"	"DELOBJ"
$ call compile "NEWSCLASSES"	"DELOBJ"
$ call compile "NEWSCONFIGS"	"DELOBJ"
$ call compile "NEWSCONTROL"	"DELOBJ"
$ call compile "NEWSDELETE"     "DELOBJ"
$ call compile "NEWSDIR"        "DELOBJ"
$ call compile "NEWSDISPLAY"    "DELOBJ"
$ call compile "NEWSDIST"       "NODELOBJ"
$ call compile "NEWSEXTRACT"    "DELOBJ"
$ call compile "NEWSFILES"      "DELOBJ"
$ call compile "NEWSFORWARD"    "DELOBJ"
$ call compile "NEWSHELP"       "DELOBJ"
$ call compile "NEWSKILL"	"DELOBJ"
$ call compile "NEWSLOCK"	"NODELOBJ"
$ call compile "NEWSVMSMAIL"    "DELOBJ"
$ call compile "NEWSMOD"        "DELOBJ"
$ call compile "NEWSMENU"       "DELOBJ"
$ call compile "NEWSMODE"       "DELOBJ"
$ call compile "NEWSPOST"       "DELOBJ"
$ call compile "NEWSPROFILE"	"DELOBJ"
$ call compile "NEWSRC"         "DELOBJ"
$ call compile "NEWSREAD"       "DELOBJ"
$ call compile "NEWSREGISTER"   "DELOBJ"
$ call compile "NEWSREMCLIENT"  "DELOBJ"
$ call compile "NEWSSEARCH"     "DELOBJ"
$ call compile "NEWSSELECT"     "DELOBJ"
$ call compile "NEWSSETSHOW"    "DELOBJ"
$ call compile "NEWSSKIM"       "DELOBJ"
$ call compile "NEWSSKIP"       "DELOBJ"
$ call compile "NEWSSITE"       "DELOBJ"
$ call compile "NEWSUTILITY"    "DELOBJ"
$ call compile "NEWSV58_UPGRADE"    "DELOBJ"
$ call compile "NEWSV60_UPGRADE"    "DELOBJ"
$ call compile "ADD_TRANSFORM"  "NODELOBJ"
$ call compile "NEWSRTL"        "NODELOBJ"
$ call compile "NEWSCMD.CLD"    "DELOBJ"
$ call compile "NEWSMSG.MSG"    "NODELOBJ"
$!
$!  NEWS
$!
$ IF make_nodebug .and. f$search("[-.news_src]news.c") .nes. "" .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   else
$     IF f$search("NEWS.OBJ").eqs.""
$     THEN
$       iftest = 1
$     else
$       objtime = f$cvtime(f$file_attributes("NEWS.OBJ","rdt"))
$       iftest =      vartime.gts.objtime -
                 .or. exttime.gts.objtime -
                 .or. inctime.gts.objtime -
                 .or. deftime.gts.objtime -
                 .or. inctime.gts.objtime -
                 .or. sittime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWS.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc news"
$     cc/nolis/object=[-.news_build]news news
$     set def [-.news_build]
$     if purge_objs then purge news.obj
$   ENDIF
$ ENDIF
$!
$ IF (f$search("NEWS.OBJ") .nes. "") .and. (f$search("NEWS_LIBRARY.OLB") .nes. "") .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NEWS.EXE") .eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest =     f$cvtime(f$file_attributes("NEWS_LIBRARY.OLB","rdt")).gts.-
                      f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWS.EXE","rdt")) -
                .or. f$cvtime(f$file_attributes("NEWS.OBJ","rdt")).gts.-
                      f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWS.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link news,news_library/library"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]news.exe news,news_library/library,link_opt/opt
$     if purge_exes then purge [-.news_dist]news.exe
$   ENDIF
$ ENDIF
$!
$!  NEWSDBG
$!
$ IF make_debug .and. f$search("[-.news_src]news.c") .nes. "" .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NEWSDBG.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("NEWSDBG.OBJ","rdt"))
$       iftest =      vartime.gts.objtime -
                 .or. exttime.gts.objtime -
		.or. inctime.gts.objtime -
		.or. deftime.gts.objtime -
		.or. inctime.gts.objtime -
		.or. sittime.gts.objtime -
		.or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWS.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc/debug/noopt news"
$     cc/nolis/object=[-.news_build]newsdbg/debug/noopt news
$     set def [-.news_build]
$     if purge_objs then purge newsdbg.obj
$   ENDIF
$ ENDIF
$ IF (f$search("NEWSDBG.OBJ") .nes. "") .and. (f$search("NEWS_LIBRARYD.OLB") .nes. "") .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NEWSDBG.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest =     f$cvtime(f$file_attributes("NEWS_LIBRARYD.OLB","rdt")).gts.-
                      f$cvtime(f$file_attributes("NEWSDBG.EXE","rdt")) -
                .or. f$cvtime(f$file_attributes("NEWSDBG.OBJ","rdt")).gts.-
                      f$cvtime(f$file_attributes("NEWSDBG.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link/debug newsdbg,news_libraryd/library"
$     link/debug/nomap newsdbg,news_libraryd/library,link_opt/opt
$     if purge_exes then purge newsdbg.exe
$   ENDIF
$ ENDIF
$!
$!  NNTP_SERVER
$!
$ call makeobj "CACHE" "CC"
$ call makeobj "CACHEM" "CC"
$ call makeobj "CACHE_LOGGER" "CC"
$ call makeobj "CACHE_VERIFY" "CC"
$ call makeobj "CACHE_HASH" "MACRO"
$ call makeobj "CACHE_MAP" "MACRO"
$ if axpflag
$ then
$   link/notrace/exe=[-.news_dist]cachem.exe -
	cachem,cache,cache_hash,cache_map,cache_verify,cache_logger,-
	[-.news_src]options_c_link_axp/opt
$ else
$   link/notrace/exe=[-.news_dist]cachem.exe -
	cachem,cache,cache_hash,cache_map,cache_verify,cache_logger,-
	[-.news_src]options_c_link/opt
$ endif
$ if purge_exes then purge [-.news_dist]cachem.exe
$ IF f$search("[-.news_src]nntp_server.c") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_SERVER.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSDEFINE.H","rdt")).gts.-
		  f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")) -
		.or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_SERVER.C","rdt")).gts.-
		  f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_server"
$     cc/nolis/object=[-.news_build]nntp_server nntp_server
$     cc/nolis/object=[-.news_build]nntp_serverd/debug/noopt nntp_server
$     set def [-.news_build]
$     if purge_objs then purge nntp_server.obj,nntp_serverd.obj
$   ENDIF
$ ENDIF
$!
$!  NNTP_TTY
$!
$ IF make_debug .and. f$search("[-.news_src]nntp_tty.c") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_TTY.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_TTY.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_TTY.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_tty"
$     cc/nolis/object=[-.news_build]nntp_tty/debug/noopt nntp_tty
$     set def [-.news_build]
$     if purge_objs then purge nntp_tty.obj
$   ENDIF
$ ENDIF
$!
$ IF make_debug .and. (f$search("NNTP_TTY.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVERD.OBJ") .nes. "")  .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_BUILD]NNTP_TTY.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_TTY.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_BUILD]NNTP_TTY.EXE","rdt")) -
		.or. f$cvtime(f$file_attributes("NNTP_SERVERD.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_BUILD]NNTP_TTY.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link/debug nntp_tty,nntp_server"
$     link/nomap/debug/exe=[-.NEWS_BUILD] -
	  nntp_tty,nntp_serverd,-
        cached,cache_mapd,cache_hashd,cache_loggerd,-
        news_libraryd/lib,link_opt/opt
$     if purge_exes then purge [-.news_build]nntp_tty.exe
$   ENDIF
$ ENDIF
$!
$!  NNTP_DEC
$!
$ IF f$search("[-.news_src]nntp_dec.c") .nes. "" .and. .not. client_only  .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_DEC.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_DEC.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_DEC.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_dec"
$     cc/nolis/object=[-.news_build]nntp_dec nntp_dec
$     set def [-.news_build]
$     if purge_objs then purge nntp_dec.obj
$   ENDIF
$ ENDIF
$!
$ IF (f$search("NNTP_DEC.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVER.OBJ") .nes. "")  .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NNTP_DEC.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_DEC.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_DEC.EXE","rdt")) -
		.or. f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_DEC.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link nntp_dec,nntp_server"
$     link/nomap/notrace/exe=[-.news_dist]nntp_dec -
	nntp_dec,nntp_server,newsdist,-
        add_transform,newsrtl,newsmsg,newslock,-
        cache,cache_map,cache_hash,cache_logger,-
        link_opt/opt
$     if purge_exes then purge [-.news_dist]nntp_dec.exe
$   ENDIF
$ ENDIF
$!
$!  NNTP_DECM
$!
$ IF f$search("[-.news_src]nntp_decm.c") .nes. "" .and. .not. client_only  .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_DECM.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_DECM.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_DECM.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_decm"
$     cc/nolis/object=[-.news_build]nntp_decm nntp_decm
$     set def [-.news_build]
$     if purge_objs then purge nntp_decm.obj
$   ENDIF
$ ENDIF
$!
$ IF (f$search("NNTP_DECM.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVER.OBJ") .nes. "") .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NNTP_DECM.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_DECM.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_DECM.EXE","rdt")) -
		.or. f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_DECM.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link nntp_decm,nntp_server"
$     link/nomap/exe=[-.NEWS_DIST]nntp_decm -
	nntp_decm,nntp_server,newsdist,-
        add_transform,newsrtl,newsmsg,newslock,-
        cache,cache_map,cache_hash,cache_logger,-
        link_opt/opt
$     if purge_exes then purge [-.news_dist]nntp_decm.exe
$   ENDIF
$ ENDIF
$!
$!  NNTP_CLIENT
$!
$ IF f$search("[-.news_src]nntp_client.c") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_CLIENT.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_CLIENT.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_CLIENT.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_client"
$     cc/nolis/object=[-.news_build]nntp_client nntp_client
$     set def [-.news_build]
$     if purge_objs then purge nntp_client.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NNTP_CLIENT.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NNTP_CLIENT.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_CLIENT.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_CLIENT.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link nntp_client"
$     link/nomap/exe=[-.NEWS_DIST]nntp_client nntp_client,newsmsg,-
        link_opt/opt
$     if purge_exes then purge [-.news_dist]nntp_client.exe
$   ENDIF
$ ENDIF
$!
$! NNTP_XMIT
$!
$ IF f$search("[-.news_src]nntpxmitcmd.cld") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTPXMITCMD.OBJ") .eqs. ""
$     THEN
$      iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTPXMITCMD.CLD","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTPXMITCMD.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ set command/obj nntpxmitcmd.cld"
$     set command/obj=[-.news_build]nntpxmitcmd.obj nntpxmitcmd.cld
$     set def [-.news_build]
$     if purge_objs then purge nntpxmitcmd.obj
$   ENDIF
$ ENDIF
$ IF f$search("[-.news_src]nntp_xmit.c") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_xmit.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_xmit.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_xmit.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_xmit"
$     cc/nolis/object=[-.news_build]nntp_xmit nntp_xmit
$     set def [-.news_build]
$     if purge_objs then purge nntp_xmit.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NNTP_XMIT.OBJ") .nes. ""  .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NNTP_XMIT.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_XMIT.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_XMIT.EXE","rdt")) -
           .or. f$cvtime(f$file_attributes("NNTPXMITCMD.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_XFER.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link nntp_xmit"
$     link/notrace/nomap/exe=[-.NEWS_DIST]NNTP_XMIT -
	nntp_xmit,nntpxmitcmd,newslock,newsrtl,newsmsg,-
	link_opt/opt
$     if purge_exes then purge [-.news_dist]nntp_xmit.exe
$   ENDIF
$ ENDIF
$!
$! NNTP_XFER
$!
$ IF f$search("[-.news_src]nntpxfercmd.cld") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTPXFERCMD.OBJ") .eqs. ""
$     THEN
$      iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTPXFERCMD.CLD","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTPXFERCMD.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ set command/obj nntpxfercmd.cld"
$     set command/obj=[-.news_build]nntpxfercmd.obj nntpxfercmd.cld
$     set def [-.news_build]
$     if purge_objs then purge nntpxfercmd.obj
$   ENDIF
$ ENDIF
$ IF f$search("[-.news_src]nntp_xfer.c") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_xfer.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_xfer.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("NNTP_xfer.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_xfer"
$     cc/nolis/object=[-.news_build]nntp_xfer nntp_xfer
$     set def [-.news_build]
$     if purge_objs then purge nntp_xfer.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NNTP_XFER.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NNTP_XFER.EXE").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NNTP_XFER.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_XFER.EXE","rdt")) -
           .or. f$cvtime(f$file_attributes("NNTPXFERCMD.OBJ","rdt")).gts.-
		  f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_XFER.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link nntp_xfer"
$     link/notrace/nomap/exe=[-.NEWS_DIST]nntp_xfer -
	nntp_xfer,nntpxfercmd,newslock,newsrtl,newsmsg,-
	link_opt/opt
$     if purge_exes then purge [-.news_dist]nntp_xfer.exe
$   ENDIF
$ ENDIF
$!
$!  NNTP_TCPCMU
$!
$ IF (use_cmu) .and. (f$search("[-.NEWS_SRC]NNTP_TCPCMU.C") .nes. "") -
 	 .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_TCPCMU.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_TCPCMU.C","rdt")).gts.-
	          f$cvtime(f$file_attributes("NNTP_TCPCMU.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_tcpcmu"
$     cc/nolis/object=[-.news_build]nntp_tcpcmu nntp_tcpcmu
$     set def [-.news_build]
$     if purge_objs then purge nntp_tcpcmu.obj
$   ENDIF
$!
$   IF (f$search("NNTP_TCPCMU.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVER.OBJ") .nes. "") .and. .not. skip_link
$   THEN
$     IF force_link
$     THEN
$       iftest = 1
$     ELSE
$       IF f$search("[-.NEWS_DIST]NNTP_TCPCMU.EXE").eqs.""
$       THEN
$         iftest = 1
$       ELSE
$         iftest =     f$cvtime(f$file_attributes("NNTP_TCPCMU.OBJ","rdt")).gts.-
	                 f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPCMU.EXE","rdt")) -
                  .or. f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")).gts.-
                        f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPCMU.EXE","rdt"))
$       ENDIF
$     ENDIF
$     IF iftest
$     THEN
$       write sys$output "$ link nntp_tcpcmu,nntp_server"
$       link/nomap/notrace/exe=[-.NEWS_DIST]nntp_tcpcmu -
	  nntp_tcpcmu,nntp_server,newsdist,-
          add_transform,newsrtl,newsmsg,newslock,-
          cache,cache_map,cache_hash,cache_logger,-
	  link_opt/opt
$       if purge_exes then purge [-.news_dist]nntp_tcpcmu.exe
$     ENDIF
$   ENDIF
$ ENDIF
$!
$! NNTP_TCPWINMULTINET
$!
$ IF (use_twg .or. use_multinet .or. use_tcpware) .and. (f$search("[-.NEWS_SRC]NNTP_TCPWINMULTINET.C") .nes. "") -
 	      .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_TCPWINMULTINET.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_TCPWINMULTINET.C","rdt")).gts.-
	          f$cvtime(f$file_attributes("NNTP_TCPWINMULTINET.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_tcpwinmultinet"
$     cc/nolis/object=[-.news_build]nntp_tcpwinmultinet nntp_tcpwinmultinet
$     set def [-.news_build]
$     if purge_objs then purge nntp_tcpwinmultinet.obj
$   ENDIF
$!
$   IF (f$search("NNTP_TCPWINMULTINET.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVER.OBJ") .nes. "") .and. .not. skip_link
$   THEN
$     IF force_link
$     THEN
$       iftest = 1
$     ELSE
$       IF f$search("[-.NEWS_DIST]NNTP_TCPWINMULTINET.EXE").eqs.""
$       THEN
$         iftest = 1
$       ELSE
$         iftest =     f$cvtime(f$file_attributes("NNTP_TCPWINMULTINET.OBJ","rdt")).gts.-
	                 f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPWINMULTINET.EXE","rdt")) -
                  .or. f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")).gts.-
                        f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPWINMULTINET.EXE","rdt"))
$       ENDIF
$     ENDIF
$     IF iftest
$     THEN
$       write sys$output "$ link nntp_tcpwinmultinet,nntp_server"
$       link/nomap/notrace/exe=[-.news_dist]nntp_tcpwinmultinet -
          nntp_tcpwinmultinet,nntp_server,newsdist,-
          add_transform,newsrtl,newsmsg,newslock,-
          cache,cache_map,cache_hash,cache_logger,-
	  link_opt/opt
$       if purge_exes then purge [-.news_dist]nntp_tcpwinmultinet.exe
$     ENDIF
$   ENDIF
$ ENDIF
$!
$! NNTP_TCPUCXM
$!
$!IF (use_ucxqios .or. use_multinet) -
$ IF (use_ucxqios) -
     .and. (f$search("[-.NEWS_SRC]NNTP_TCPUCXM.C") .nes. "") .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile 
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NNTP_TCPUCXM.OBJ").eqs.""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NNTP_TCPUCXM.C","rdt")).gts.-
	          f$cvtime(f$file_attributes("NNTP_TCPUCXM.OBJ","rdt")) 
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc nntp_tcpucxm"
$     cc/nolis/object=[-.news_build]nntp_tcpucxm nntp_tcpucxm
$     set def [-.news_build]
$     if purge_objs then purge nntp_tcpucxm.obj
$   ENDIF
$!
$   IF (f$search("NNTP_TCPUCXM.OBJ") .nes. "") .and. -
       (f$search("NNTP_SERVER.OBJ") .nes. "") .and. .not. skip_link
$   THEN
$     IF force_link
$     THEN
$       iftest = 1
$     ELSE
$       IF f$search("[-.NEWS_DIST]NNTP_TCPUCXM.EXE").eqs.""
$       THEN
$         iftest = 1
$       ELSE
$         iftest =     f$cvtime(f$file_attributes("NNTP_TCPUCXM.OBJ","rdt")).gts.-
	                 f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPUCXM.EXE","rdt")) -
                  .or. f$cvtime(f$file_attributes("NNTP_SERVER.OBJ","rdt")).gts.-
                        f$cvtime(f$file_attributes("[-.NEWS_DIST]NNTP_TCPUCXM.EXE","rdt")) 
$       ENDIF
$     ENDIF
$     IF iftest
$     THEN
$       write sys$output "$ link nntp_tcpucxm,nntp_server"
$       link/nomap/exe=[-.news_dist]nntp_tcpucxm -
          nntp_tcpucxm,nntp_server,newsdist,-
          add_transform,newsrtl,newsmsg,newslock,-
          cache,cache_map,cache_hash,cache_logger,-
	  link_opt/opt
$       if purge_exes then purge [-.news_dist]nntp_tcpucxm.exe
$     ENDIF
$   ENDIF
$ ENDIF
$!
$!  NEWS.HLP
$!
$ IF f$search("[-.NEWS_SRC]NEWS.HLP") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NEWS.HLB").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWS.HLP","rdt")).gts.-
                  f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWS.HLB","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ libr/help/create news news"
$     set def [-.news_dist]
$     libr/help/create news [-.news_src]news
$     set def [-.news_build]
$     if purge_exes then purge [-.news_dist]news.hlb
$   ENDIF
$ ENDIF
$!
$!  ARBITRON.EXE - same for arbitron
$!
$ IF f$search("[-.NEWS_SRC]ARBITRON.C") .nes. "" .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("ARBITRON.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("[-.NEWS_SRC]ARBITRON.C","rdt")).gts.-
                  f$cvtime(f$file_attributes("ARBITRON.OBJ","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc arbitron"
$     cc/nolis/opt/object=[-.news_build]arbitron arbitron
$     set def [-.news_build]
$     if purge_objs then purge arbitron.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("ARBITRON.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]ARBITRON.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("ARBITRON.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]ARBITRON.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link arbitron"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]arbitron.exe -
	arbitron,news_library/lib,link_opt/opt
$     if purge_exes then purge [-.news_dist]arbitron.exe
$   ENDIF
$ ENDIF
$!
$!  NEWITMFILE.EXE
$!
$ IF f$search("[-.NEWS_SRC]NEWITMFILE.C") .nes. ""  .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NEWITMFILE.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("NEWITMFILE.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWITMFILE.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc newitmfile"
$     cc/nolis/object=[-.news_build]NEWITMFILE NEWITMFILE
$     set def [-.news_build]
$     if purge_objs then purge NEWITMFILE.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NEWITMFILE.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NEWITMFILE.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NEWITMFILE.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWITMFILE.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link newitmfile"
$     link/nomap/exe=[-.news_dist]NEWITMFILE.exe -
	NEWITMFILE,-
	news_library/lib,link_opt/opt
$     if purge_exes then purge [-.news_dist]NEWITMFILE.exe
$   ENDIF
$ ENDIF
$!
$!  NEWGRPFILE.EXE
$!
$ IF f$search("[-.NEWS_SRC]NEWGRPFILE.C") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NEWGRPFILE.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("NEWGRPFILE.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWGRPFILE.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc newgrpfile"
$     cc/nolis/object=[-.news_build]NEWGRPFILE NEWGRPFILE
$     set def [-.news_build]
$     if purge_objs then purge NEWGRPFILE.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NEWGRPFILE.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NEWGRPFILE.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NEWGRPFILE.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWGRPFILE.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link newgrpfile"
$     link/nomap/exe=[-.news_dist]NEWGRPFILE.exe -
	NEWGRPFILE,-
	news_library/lib,link_opt/opt
$     if purge_exes then purge [-.news_dist]NEWGRPFILE.exe
$   ENDIF
$ ENDIF
$!
$!  NEWSSHUTDOWN.EXE
$!
$ IF f$search("[-.NEWS_SRC]NEWSSHUTDOWN.C") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("NEWSSHUTDOWN.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("NEWSSHUTDOWN.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]NEWSSHUTDOWN.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc newsshutdown"
$     cc/nolis/object=[-.news_build]NEWSSHUTDOWN NEWSSHUTDOWN
$     set def [-.news_build]
$     if purge_objs then purge NEWSSHUTDOWN.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("NEWSSHUTDOWN.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]NEWSSHUTDOWN.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("NEWSSHUTDOWN.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]NEWSSHUTDOWN.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link newsshutdown"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]NEWSSHUTDOWN.exe -
	NEWSSHUTDOWN, -
	news_library/lib,link_opt/opt
$     if purge_exes then purge [-.news_dist]NEWSSHUTDOWN.exe
$   ENDIF
$ ENDIF
$!
$!  FEEDCHECK.EXE
$!
$ IF f$search("[-.NEWS_SRC]FEEDCHECK.C") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("FEEDCHECK.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("FEEDCHECK.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]FEEDCHECK.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc feedcheck"
$     cc/nolis/object=[-.news_build]FEEDCHECK FEEDCHECK
$     set def [-.news_build]
$     if purge_objs then purge FEEDCHECK.obj
$   ENDIF
$ ENDIF
$ IF f$search("[-.NEWS_SRC]FEEDCHECKCMD.CLD") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("FEEDCHECKCMD.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("FEEDCHECKCMD.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]FEEDCHECKCMD.CLD","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ set command/obj feedcheckcmd"
$     set command/object=[-.news_build]FEEDCHECKCMD FEEDCHECKCMD
$     set def [-.news_build]
$     if purge_objs then purge FEEDCHECKCMD.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("FEEDCHECK.OBJ") .nes. "" .and. f$search("FEEDCHECKCMD.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]FEEDCHECK.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("FEEDCHECK.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]FEEDCHECK.EXE","rdt")).or.-
		 f$cvtime(f$file_attributes("FEEDCHECKCMD.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]FEEDCHECK.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link feedcheck"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]FEEDCHECK.exe	-
	FEEDCHECK,FEEDCHECKCMD,-
	link_opt/opt
$     if purge_exes then purge [-.news_dist]FEEDCHECK.exe
$   ENDIF
$ ENDIF
$!
$!  BATCHNEWS.EXE
$!
$ IF f$search("[-.NEWS_SRC]BATCHNEWS.C") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("BATCHNEWS.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("BATCHNEWS.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]BATCHNEWS.C","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ cc batchnews"
$     cc/nolis/object=[-.news_build]BATCHNEWS BATCHNEWS
$     set def [-.news_build]
$     if purge_objs then purge BATCHNEWS.obj
$   ENDIF
$ ENDIF
$ IF f$search("[-.NEWS_SRC]BATCHNEWSCMD.CLD") .nes. "" .and. .not. client_only .and. .not. skip_compile
$ THEN
$   IF force_compile
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("BATCHNEWSCMD.OBJ").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       objtime = f$cvtime(f$file_attributes("BATCHNEWSCMD.OBJ","rdt"))
$       iftest =      deftime.gts.objtime -
                 .or. f$cvtime(f$file_attributes("[-.NEWS_SRC]BATCHNEWSCMD.CLD","rdt")).gts.objtime
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     set def [-.news_src]
$     write sys$output "$ set command/obj batchnewscmd"
$     set command/object=[-.news_build]BATCHNEWSCMD BATCHNEWSCMD
$     set def [-.news_build]
$     if purge_objs then purge BATCHNEWSCMD.obj
$   ENDIF
$ ENDIF
$!
$ IF f$search("BATCHNEWS.OBJ") .nes. "" .and. f$search("BATCHNEWSCMD.OBJ") .nes. "" .and. .not. skip_link
$ THEN
$   IF force_link
$   THEN
$     iftest = 1
$   ELSE
$     IF f$search("[-.NEWS_DIST]BATCHNEWS.EXE").eqs. ""
$     THEN
$       iftest = 1
$     ELSE
$       iftest = f$cvtime(f$file_attributes("BATCHNEWS.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]BATCHNEWS.EXE","rdt")).or.-
		 f$cvtime(f$file_attributes("BATCHNEWSCMD.OBJ","rdt")).gts.-
	     f$cvtime(f$file_attributes("[-.NEWS_DIST]BATCHNEWS.EXE","rdt"))
$     ENDIF
$   ENDIF
$   IF iftest
$   THEN
$     write sys$output "$ link batchnews"
$     link/notrace/nodebug/nomap/exe=[-.news_dist]BATCHNEWS.exe -
	BATCHNEWS,BATCHNEWSCMD,-
	link_opt/opt
$     if purge_exes then purge [-.news_dist]BATCHNEWS.exe
$   ENDIF
$ ENDIF
$ write sys$output "NEWS FILES Built"
$done:
$ set def [-.news_src]
$ if newsrclog .nes. "" then $ define newsrc 'newsrclog
$ exit ! 'f$verify(makehay_ver)'
$!
$!  COMPILE SUBROUTINE
$!
$compile: subroutine
$   on control_y THEN $ stop
$   IF skip_compile THEN $  goto compdun
$   srcname = f$parse(p1,"[-.news_src].c")
$   srctype = f$parse(srcname,,,"TYPE") - "."
$   objname = f$parse(".obj;",p1) - ";"
$   objcomp = f$parse("[-.news_build].obj;",p1) - ";"
$   purgename = f$parse(p1,,,"NAME") + "*.*," +-
      "[-.news_src]" + f$parse(p1,,,"NAME") + "*.*"
$   srctime = f$cvtime(f$file_attributes(srcname,"rdt"))
$   IF p2 .eqs. "NODELOBJ" THEN $ goto testcomp
$   IF make_nodebug THEN goto testcomp
$   goto dbgcomp
$    testcomp:
$     IF force_compile THEN $ goto ccit
$     IF f$search(objname).eqs."" THEN $ goto ccit
$     objtime = f$cvtime(f$file_attributes(objname,"rdt"))
$     IF srctime.gts.objtime THEN $ goto ccit
$     IF vartime.gts.objtime THEN $ goto ccit
$     IF exttime.gts.objtime THEN $ goto ccit
$     IF deftime.gts.objtime THEN $ goto ccit
$     IF inctime.gts.objtime THEN $ goto ccit
$     IF sittime.gts.objtime THEN $ goto ccit
$     goto dbgcomp
$    ccit:
$     set def [-.news_src]
$     goto cmp_'srctype'
$     goto cmp_UNKNOWN
$cmp_C:
$     write sys$output "$ cc/nodebug/opt ''p1'"
$     cc/nolis/nodebug/opt/object='objcomp 'srcname
$     goto cmp_endcase
$cmp_MAR:
$     write sys$output "$ macro ''p1'"
$     macro/nolis/nodebug/object='objcomp 'srcname
$     goto cmp_endcase
$cmp_CLD:
$     write sys$output "$ set command/obj ''p1'"
$     set command/obj='objcomp 'srcname
$     goto cmp_endcase
$cmp_MSG:
$     write sys$output "$ message ''p1'"
$     message/nolis/object='objcomp 'srcname
$     goto cmp_endcase
$cmp_UNKNOWN:
$     write sys$output "unknown source file type: ",srctype
$     goto cmp_endcase
$cmp_endcase:
$     set def [-.news_build]
$     library/object news_library 'objname
$     If .not.purge_objs Then goto dbgcomp
$     IF p2 .eqs. "NODELOBJ" THEN $ goto purgeit
$     create 'objname
	COMPILE MARKER
$    purgeit:
$     purge 'objname
$ dbgcomp:
$   IF .not. make_debug THEN $  goto compdun
$    dbgchk:
$     objname = f$parse(f$parse(p1,,,"NAME")+"d", ".obj;") - ";"
$     objcomp = f$parse(f$parse(p1,,,"NAME")+"d", "[-.news_build].obj;") - ";"
$     IF force_compile THEN $ goto dbgccit
$     IF f$search(objname).eqs."" THEN $ goto dbgccit
$     objtime = f$cvtime(f$file_attributes(objname,"rdt"))
$     IF srctime.gts.objtime THEN $ goto dbgccit
$     IF vartime.gts.objtime THEN $ goto dbgccit
$     IF exttime.gts.objtime THEN $ goto dbgccit
$     IF deftime.gts.objtime THEN $ goto dbgccit
$     IF inctime.gts.objtime THEN $ goto dbgccit
$     IF sittime.gts.objtime THEN $ goto dbgccit
$     goto compdun
$    dbgccit:
$     set def [-.news_src]
$     goto cmpdbg_'srctype'
$     goto cmpdbg_UNKNOWN
$cmpdbg_C:
$     write sys$output "$ cc/debug/noopt ''p1'"
$     cc/nolis/debug/noopt/object='objcomp 'srcname
$     goto cmpdbg_endcase
$cmpdbg_MAR:
$     write sys$output "$ macro/debug ''p1'"
$     macro/nolis/debug/object='objcomp 'srcname
$     goto cmpdbg_endcase
$cmpdbg_CLD:
$     write sys$output "$ set command/obj ''p1'"
$     set command/obj='objcomp 'srcname
$     goto cmpdbg_endcase
$cmpdbg_MSG:
$     write sys$output "$ message ''p1'"
$     message/nolis/object='objcomp 'srcname
$     goto cmpdbg_endcase
$cmpdbg_UNKNOWN:
$     write sys$output "unknown source file type: ",srctype
$     goto cmpdbg_endcase
$cmpdbg_endcase:
$     set def [-.news_build]
$     library/object news_libraryd 'objname
$     If .not.purge_objs Then goto compdun
$     create 'objname
	COMPILE MARKER
$     purge 'objname
$ compdun:
$   exit
$ endsubroutine
$!
$!  MAKEOBJ SUBROUTINE
$!
$makeobj: subroutine
$   on control_y THEN $ stop
$   IF skip_compile THEN $  goto makedun
$   extn = ".C"
$   if p2 .eqs. "MACRO" THEN $ extn = ".MAR" 
$   srcname = "[-.news_src]" + p1 + extn
$   objname = p1 + ".obj"
$   objcomp = "[-.news_build]" + p1
$   purgename = p1 + "*.*, [-.news_src]" + p1 + "*.*"
$   srctime = f$cvtime(f$file_attributes(srcname,"rdt"))
$   IF force_compile THEN $ goto ccit
$   IF f$search(objname).eqs."" THEN $ goto ccit
$   objtime = f$cvtime(f$file_attributes(objname,"rdt"))
$   IF srctime.gts.objtime THEN $ goto ccit
$   goto makedun
$   ccit:
$     set def [-.news_src]
$     write sys$output "$ ''p2'/nodebug ''p1'"
$     if p2 .eqs. "CC" 
$	THEN
$	  if make_nodebug then cc/nolis/nodebug/opt/object='objcomp' 'p1'
$	  if make_debug then cc/nolis/debug/noopt/object='objcomp'd 'p1'
$     endif
$     if p2 .eqs. "MACRO"
$	THEN
$	  if axpflag
$	  then
$	    if make_nodebug then macro/nolis/nodebug/migration/unalign/object='objcomp' 'p1'
$	    if make_debug then macro/nolis/debug/migration/unalign/object='objcomp'd 'p1'
$	  else
$	    if make_nodebug then macro/nolis/nodebug/object='objcomp' 'p1'
$	    if make_debug then macro/nolis/debug/object='objcomp'd 'p1'
$	endif
$     endif
$     set def [-.news_build]
$     if purge_objs.and.make_nodebug then purge 'objcomp'.obj
$     if purge_objs.and.make_debug then purge 'objcomp'd.obj
$ makedun:
$   exit
$ endsubroutine
